"""
Metadata API

Metadata Coverage

.. _Metadata Coverage:
   https://developer.salesforce.com/docs/metadata-coverage/54
"""

import threading
import sys
# import xml.etree.ElementTree as ET
import requests

from zeep import Client, Settings

from SFDCFW.Constant import SFDC_API_V

class Metadata:
    """Metadata class."""

    def __init__(self, access, wsdl):
        """Constructor

        Args:
            access (tuple): The Salesforce session ID / access token and
                server URL / instance URL tuple
            wsdl (str): The path to the WSDL file
        """

        # Unpack the tuple for session ID / access token and server URL / instance URL
        self.id_token, self.url = access

        # Create client with setting of disable strict mode, use recovery mode
        setting = Settings(strict=False)
        client = Client(wsdl, settings=setting)
        # Create the service with custom binding and URL
        binding = "{http://soap.sforce.com/2006/04/metadata}MetadataBinding"
        self.service = client.create_service(binding, self.url)

        # Create the SOAP header (this is different than the HTTP header)
        self.soap_header = {
            "SessionHeader": {
                "sessionId": self.id_token
            }
        }


    def list_metadata(self, query):
        """List Metadata

        Args:
            query (list): A complex list of dictionary specify components
                [
                    { "folder": "Report", "type": "ReportName" },
                    { "type": "WorkflowRule" }
                ]

        Returns:
            A complex list of dictionary for the requested metadata
            component(s)
        """

        return self.service.listMetadata(queries=query,
                                         asOfVersion=SFDC_API_V,
                                         _soapheaders=self.soap_header)


    def read_metadata(self, metadata_type, full_name, thread=32):
        """Read Metadata

        Args:
            metadata_type (str): The type of the metadata
            full_name (list): A complex list of string for the full
                name(s) of the metadata component

        Returns:
            A complex list of dictionary for the requested metadata
            component(s)
        """

        # Create a new list to store all the read result
        read_result_all = []
        # Get the size of the full name list
        full_name_size = len(full_name)
        # Limit for the metadata API function
        record_limit = 10

        # Loop through the full name list 10 records at a time
        for i in range(0, full_name_size, record_limit):
            # Make request(s) to read metadata, 10 at a time
            read_result = self.service.readMetadata(type=metadata_type,
                                                    fullNames=full_name[ i : i + record_limit ],
                                                    _soapheaders=self.soap_header)
            # Add the current read result to all the read result
            read_result_all.extend(read_result)

        return read_result_all


    def read_metadata_thread(self, metadata_type, full_name, thread=10):
        """Read Metadata

        Args:
            metadata_type (str): The type of the metadata
            full_name (list): A complex list of string for the full
                name(s) of the metadata component
            thread (int): The number of thread to use

        Returns:
            A complex list of dictionary for the requested metadata
            component(s)
        """

        def _read_metadata_thread(self, metadata_type, full_name):
            """Read Metadata Thread Function

            Args:
                metadata_type (str): The type of the metadata
                full_name (list): A complex list of string for the full
                    name(s) of the metadata component
            """
            # Make request(s) to read metadata, 10 at a time
            read_result = self.service.readMetadata(type=metadata_type,
                                                    fullNames=full_name,
                                                    _soapheaders=self.soap_header)

            # Add the current read result to all the read result
            read_result_all.extend(read_result)

        # Create a new list to store all the read result
        read_result_all = []
        # Get the size of the full name list
        full_name_size = len(full_name)
        # Limit for the metadata API function
        record_limit = 10

        # Round the thread to multiple of 10
        # thread_size = thread + (-thread) % 10
        thread_size = thread
        # Calculate the thread batch size based on metadata API limit
        thread_batch_size = (full_name_size + (-full_name_size) % record_limit) // record_limit
        # Check which is smaller, thread batch size or thread size
        # Take the smaller of the two
        thread_count = thread_batch_size if thread_batch_size < thread_size else thread_size
        # Create a thread list
        t_list = []

        # Loop through the full name list
        for i in range(0, thread_batch_size, thread_size):
            # Account for last iteration, maximum thread (thread_size) not needed
            if thread_batch_size - i < thread_size:
                # Use only maximum thread needed
                thread_count = thread_batch_size - i

            # Loop through and execute the thread
            for j in range(thread_count):
                # Calculate full name start and end range
                full_name_start = (j + i) * record_limit
                # Calculate the precise end range for last iteration...
                # Because I can...
                if full_name_start + record_limit > full_name_size:
                    full_name_end = full_name_start + thread_count
                else:
                    full_name_end = full_name_start + record_limit

                # Create the thread
                t = threading.Thread(target=_read_metadata_thread,
                                     args=(self, metadata_type, full_name[ full_name_start : full_name_end ],))
                # Add it to a thread list
                t_list.append(t)
                # Start the thread
                t.start()

            # Wait for all threads to finish
            for t in t_list:
                t.join()

        return read_result_all


    def update_metadata(self, metadata):
        """Update Metadata

        Args:
            metadata (list): A complex list of dictionary for metadata
                component to update

        Returns:
            A complex list of dictionary for result of the metadata
            component update
        """

        # Create a new list to store all the update result
        update_result_all = []
        # Get the size of the metadata list
        metadata_size = len(metadata)
        # Limit for the metadata API function
        record_limit = 10

        # Loop through the metadata list 10 records at a time
        for i in range(0, metadata_size, record_limit):
            # Make request(s) to update metadata, 10 at a time
            update_result = self.service.updateMetadata(metadata=metadata[ i : i + record_limit ],
                                                        _soapheaders=self.soap_header)
            # Add the current update result to all the update result
            update_result_all.extend(update_result)

        return update_result_all

    def update_metadata_thread(self, metadata, thread=10):
        """Update Metadata

        Args:
            metadata (list): A complex list of dictionary for metadata
                component to update
            thread (int): The number of thread to use

        Returns:
            A complex list of dictionary for result of the metadata
                component update
        """

        def _update_metadata_thread(self, metadata):
            """Update Metadata

            Args:
                metadata (list): A complex list of dictionary for
                    metadata component to update

            Returns:
                A complex list of dictionary for result of the metadata
                    component update
            """

            # Make request(s) to update metadata, 10 at a time
            update_result = self.service.updateMetadata(metadata=metadata,
                                                        _soapheaders=self.soap_header)

            # Add the current update result to all the update result
            update_result_all.extend(update_result)

        # Create a new list to store all the update result
        update_result_all = []
        # Get the size of the metadata list
        metadata_size = len(metadata)
        # Limit for the metadata API function
        record_limit = 10

        # Thread size
        thread_size = thread
        # Calculate the thread batch size based on metadata API limit
        thread_batch_size = (metadata_size + (-metadata_size) % record_limit) // record_limit
        # Check which is smaller, thread batch size or thread size
        # Take the smaller of the two
        thread_count = thread_batch_size if thread_batch_size < thread_size else thread_size
        # Create a thread list
        t_list = []

        # Loop through the metadata list
        for i in range(0, thread_batch_size, thread_size):

            # Account for last iteration, maximum thread (thread_size) not needed
            if thread_batch_size - i < thread_size:
                # Use only maximum thread needed
                thread_count = thread_batch_size - i
            
            for j in range(thread_count):
                # Calculate metadata start and end range
                metadata_start = (j + i) * record_limit
                metadata_end = metadata_start + record_limit
                # Create the thread
                t = threading.Thread(target=_update_metadata_thread,
                                     args=(self, metadata[ metadata_start : metadata_end ],))
                # Add it to a thread list
                t_list.append(t)
                # Start the thread
                t.start()

            # Wait for all threads to finish
            for t in t_list:
                t.join()

        return update_result_all


"""
SFDCFW.Access
~~~~~~~~~~~~~
"""

import requests
# import xml.etree.ElementTree as ET
from zeep import Client, Settings, exceptions

# try:
#     # Python 3+
#     from html import escape
# except ImportError:
#     # Python Older
#     from cgi import escape

from SFDCFW.Constant import SFDC_API_V


class Access:
    """Access."""

    def __init__(self,
                 username,
                 password,
                 security_token,
                 client_id=None,
                 client_secret=None,
                 version=SFDC_API_V,
                 domain='login',
                 wsdl=None,
                 metadata=False):
        """Constructor

        Args:
            username (str): The Salesforce user Username
            password (str): The Salesforce user Password
            security_token (str): The Salesforce user Security Token
            client_id (str): The Salesforce Connected App Consumer Key
            client_secret (str): The Salesforce Connected App Consumer Secret
            version (str): The Salesforce version of the Application Programming Interface
            domain (str): The common Salesforce domain for connection (login or test)
            wsdl (str): The path to the WSDL (Web Services Description Language) file
            metadata (bool): Whether or not this is for metadata
        """

        self.username = username
        self.password = password
        self.security_token = security_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = None
        self.soap_url = f'https://{domain}.salesforce.com/services/Soap/u/{version}'
        self.rest_url = f'https://{domain}.salesforce.com/services/oauth2/token'
        self.wsdl = wsdl
        self.metadata = metadata


    def login(self):
        """Login

        Returns:
            A tuple containing the session ID / access token and
            (metadata) server URL / instance URL based on credential
        """

        # Check if username, password, security token, client ID, client secret is provided
        if all(credential is not None for credential in [self.username,
                                                         self.password,
                                                         self.security_token,
                                                         self.client_id,
                                                         self.client_secret]):
            # If username, password, security token, client ID, client secret is provided, login with REST
            return self.login_rest(username=self.username,
                                   password=self.password,
                                   security_token=self.security_token,
                                   client_id=self.client_id,
                                   client_secret=self.client_secret,
                                   url=self.rest_url)

        # Check if username, password is provide
        elif all(credential is not None for credential in [self.username,
                                                           self.password,
                                                           self.security_token]):
            # If username, password is provided, login with SOAP
            return self.login_soap(username=self.username,
                                   password=self.password,
                                   security_token=self.security_token,
                                   url=self.soap_url,
                                   wsdl=self.wsdl,
                                   metadata=self.metadata)
        
        # There is an error
        return None
    

    def login_rest(self,
                   username,
                   password,
                   security_token,
                   client_id=None,
                   client_secret=None,
                   url=None):
        """Login via REST (REpresentational State Transfer)

        Args:
            username (str): The Salesforce user username
            password (str): The Salesforce user password
            security_token (str): The Salesforce user Security Token
            client_id (str): The Salesforce Connected App Consumer Key
            client_secret (str): The Salesforce Connected App Consumer Secret
            url (str): The Salesforce REST API URL used for login

        Returns:
            A string for the Salesforce bearer token if call was
            successful, or None otherwise
        """

        # Create header
        header = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        # Create payload
        payload = {
            'grant_type': 'password',
            'username': username,
            'password': password + security_token,
            'client_id': client_id,
            'client_secret': client_secret
        }

        # Make the request and get the response
        r = requests.post(url, headers=header, data=payload)

        if r.status_code == 200:
            # Parse the access token and instance URL
            access_token = r.json()['access_token']
            instance_url = r.json()['instance_url']
            return (access_token, instance_url)

        # Return None for now if error
        return None


    def login_soap(self,
                   username,
                   password,
                   security_token,
                   url=None,
                   wsdl=None,
                   metadata=False):
        """Login via SOAP (Simple Object Access Protocol)

        Args:
            username (str): The Salesforce user username
            password (str): The Salesforce user password
            security_token (str): The Salesforce user Security Token
            url (str): The Salesforce SOAP API URL used for login
            wsdl (str): The path to the WSDL (Web Services Description Language) file
            metadata (bool): Whether or not to get the metadata server URL

        Returns:
            A tuple containing the session ID and (metadata) server URL
        """

        # Create client with setting of disable strict mode, use recovery mode
        setting = Settings(strict=False)
        client = Client(wsdl, settings=setting)

        try:
            # Attempt to make the request for the response
            r = client.service.login(username, password + security_token)
        except exceptions.Fault:
            # Return None for now if exception
            return None

        # Check to see if the client service succeeded
        if r is not None:
            # Get the sessionId
            session_id = r['sessionId']

            if metadata:
                # Get the `metadataServerUrl` if requested
                metadata_server_url = r['metadataServerUrl']

                # Return a tuple of the session ID and metadata server URL
                return (session_id, metadata_server_url)
            else:
                # Get the `serverUrl`
                server_url = r['serverUrl']

                # Return a tuple of the session ID and server URL
                return (session_id, server_url)

        # Return None for now if error
        return None


"""
Tooling API
"""

import json
from urllib.parse import urlencode

from SFDCFW.Rest.Rest import Rest
from SFDCFW.Constant.Constant import SFDC_API_V
from SFDCFW.Constant.Constant import HTTP_GET
from SFDCFW.Constant.Constant import HTTP_PATCH

class Tooling(Rest):
    """Tooling API.
    """

    def query(self, query):
        """Execute SOQL (Salesforce Object Query Language) Query

        Args:
            query (str): The SOQL (Salesforce Object Query Language) query

        Returns:
            A string formatted JSON for the query response
        """

        # Create the request URL
        request_url = "/tooling/query/?q=" + query

        # Send the request
        r = self.send(HTTP_GET, request_url, None)

        return r.text
    
    def execute_anonymous(self, anonymous_body):
        """Execute Anonymous

        Args:
            anonymous_body (str): The Apex code to be executed as an anonymous
                body

        Returns:
            A string formatted JSON for the HTTP response object
        """

        # Create the query string
        query = urlencode({'anonymousBody': anonymous_body})

        # Create the request URL
        request_url = "/tooling/executeAnonymous/?" + query

        # Send the request
        r = self.send(HTTP_GET, request_url, None)

        return r.text


"""
SFDCFW.Test.TestMetadata
~~~~~~~~~~~~~~~~~~~~~~~~
"""

import json
import os
import unittest

from SFDCFW.Access import Access
from SFDCFW.Metadata import Metadata
from SFDCFW.Constant import TEST_DATA


def setUpModule():
    """Set Up Module"""
    pass


def tearDownModule():
    """Tear Down Module"""
    pass

class TestMetadata(unittest.TestCase):
    """Test Metadata."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the Test Data from JSON (JavaScript Object Notation) file.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.domain = cls.data['domain']

        # Get the WSDL (Web Service Definition Language) file path
        cls.enterprise_wsdl = os.path.join(current_directory, cls.data['enterprise_wsdl'])
        cls.metadata_wsdl = os.path.join(current_directory, cls.data["metadata_wsdl"])

        # Get the SOAP Access user data for success login
        soap_access_user_success = cls.data['user']['soap_access_user_success']

        # Create an instance of Access object and login
        cls.access = Access(username=soap_access_user_success['username'],
                            password=soap_access_user_success['password'],
                            security_token=soap_access_user_success['security_token'],
                            domain=cls.domain,
                            wsdl=cls.enterprise_wsdl,
                            metadata=True).login()


    def test_metadata_general(self):
        metadata = Metadata(access=self.access,
                            wsdl=self.metadata_wsdl)


import csv
#
from SFDCFW.Soap.Metadata import Metadata

class CustomField(Metadata):
    """Custom Field.
    """

    def __init__(self, access, wsdl):
        """Constructor.

        Args:
            access (tuple): The Salesforce session ID / access token and server
                URL / instance URL tuple.
            wsdl (str): The path to the WSDL file.
        """

        # Use the Metadata API
        self.metadata = Metadata(access, wsdl)

    def toggle_lookup_filter_parallel(self, active, thread=20, process=10):
        """(De)Activate Lookup Filter(s) Parallel.

        Args:
            active (bool): A flag determine whether to activate or deactivate
                the Lookup Filter(s).
            thread (int): The number of thread to use
            process (int): The number of process to use

        Returns:
            A complex list of dictionary for result of the Lookup Filter(s)
                (de)activation.
        """

        # Create the query
        query = [{"type": "CustomField"}]

        # Get the list of Custom Field(s)
        custom_field_list = self.metadata.list_metadata(query)

        # Parse a list of full name from the Custom Field(s) result
        full_name_result = [custom_field["fullName"] for custom_field in custom_field_list]

        # Get the detail read result information from the full name
        read_result = self.metadata.read_metadata_thread(query[0]["type"], full_name_result, thread=thread)

        # Slice out the Custom Field(s) result that have Lookup Filter(s)
        # Look for those that are (in)active
        if active:
            lookup_filter_list = [record for record in read_result if record["fullName"] and record["lookupFilter"] and not record["lookupFilter"]["active"]]
        else:
            lookup_filter_list = [record for record in read_result if record["fullName"] and record["lookupFilter"] and record["lookupFilter"]["active"]]

        # Loop through the result and update the active status
        for record in lookup_filter_list:
            record["lookupFilter"]["active"] = active

        a = {
            'fullName': 'AcctPlan_White_Space__c.Account_BU__c',
            'businessOwnerGroup': None,
            'businessOwnerUser': None,
            'businessStatus': None,
            'caseSensitive': None,
            'customDataType': None,
            'defaultValue': None,
            'deleteConstraint': 'Restrict',
            'deprecated': None,
            'description': 'Account/BU that this White Space record is related to.',
            'displayFormat': None,
            'encryptionScheme': None,
            'escapeMarkup': None,
            'externalDeveloperName': None,
            'externalId': False,
            'fieldManageability': None,
            'formula': None,
            'formulaTreatBlanksAs': None,
            'inlineHelpText': 'Search for Account/BU entered on the Account/BU Tab',
            'isAIPredictionField': None,
            'isConvertLeadDisabled': None,
            'isFilteringDisabled': None,
            'isNameField': None,
            'isSortingDisabled': None,
            'label': 'Account/BU',
            'length': None,
            'lookupFilter': {
                'active': False,
                'booleanFilter': None,
                'description': None,
                'errorMessage': None,
                'filterItems': [
                    {
                        'field': 'Account_BU__c.AcctPlan_ParentPlan__c',
                        'operation': 'equals',
                        'value': None,
                        'valueField': '$Source.Account_Plan__c'
                    }
                ],
                'infoMessage': None,
                'isOptional': False
            },
            'maskChar': None,
            'maskType': None,
            'metadataRelationshipControllingField': None,
            'populateExistingRows': None,
            'precision': None,
            'referenceTargetField': None,
            'referenceTo': 'Account_BU__c',
            'relationshipLabel': 'Account Plan White Space',
            'relationshipName': 'AcctPlan_White_Space',
            'relationshipOrder': None,
            'reparentableMasterDetail': None,
            'required': True,
            'restrictedAdminField': None,
            'scale': None,
            'securityClassification': None,
            'startingNumber': None,
            'stripMarkup': None,
            'summarizedField': None,
            'summaryFilterItems': [],
            'summaryForeignKey': None,
            'summaryOperation': None,
            'trackFeedHistory': None,
            'trackHistory': True,
            'trackTrending': False,
            'translateData': None,
            'type': 'Lookup',
            'unique': None,
            'valueSet': None,
            'visibleLines': None,
            'writeRequiresMasterRead': None
        }

        b = { k: v for k, v in a.items() if v is not None }

        # # Update the metadata
        update_result = self.metadata.update_metadata_thread([b])

        # Return the update result
        return update_result


"""
Validation Rule
"""

import csv
from SFDCFW.Soap.Metadata import Metadata

class ValidationRule(Metadata):
    """Validation Rule.
    """

    def __init__(self, access, wsdl):
        """Constructor.

        Args:
            access (tuple): The Salesforce session ID / access token and server
                URL / instance URL tuple.
            wsdl (str): The path to the WSDL file.
        """

        # Use the Metadata API
        self.metadata = Metadata(access, wsdl)


    def get(self, active=None, full_name=None):
        """
        """
        pass


    def toggle_active(self, active, full_name=None):
        """(De)Activate specific Validation Rule(s) for a given list of full name(s).

        Args:
            active (bool): A flag determine whether to activate or deactivate
                the Validation Rule(s).
            full_name (list): A list of string(s) for the full name of the
                Validation Rule to (de)activate. Will (de)activate all
                Validation Rule(s) if no full name is passed.

        Returns:
            A complex list of dictionary for result of the Validation Rule(s)
                (de)activation.
        """

        # Create the query
        query = [{"type": "ValidationRule"}]

        # Get the list of Valiation Rule(s)
        validation_rule_list = self.metadata.list_metadata(query)
        # print(validation_rule_list)

        if full_name is not None:
            # Find full_name in validation_rule_list
            # Loop through list of Validation Rule(s) (validation_rule_list) and find the requested (full_name) element
            validation_rule_result = [validation_rule for validation_rule in validation_rule_list if validation_rule["fullName"] in full_name]
        else:
            validation_rule_result = validation_rule_list

        # Parse a list of full name from the Validation Rule(s) result
        full_name_result = [validation_rule["fullName"] for validation_rule in validation_rule_result]

        # Get the detail read result information from the full name
        read_result = self.metadata.read_metadata(query[0]["type"], full_name_result)

        # Slice out the Validation Rule(s) result that are (in)active
        if active:
            read_result[:] = [record for record in read_result if not record["active"]]
        elif not active:
            read_result[:] = [record for record in read_result if record["active"]]

        # Loop through the result and update the active status
        for record in read_result:
            if active:
                record["active"] = True
            elif not active:
                record["active"] = False

        # Update the metadata
        update_result = self.metadata.update_metadata(read_result)

        # Return the update result
        return update_result


"""
SFDCFW.SFDCFW
~~~~~~~~~~~~~
"""

from urllib.parse import urlparse

from SFDCFW.Access import Access
from SFDCFW.Rest.SObject import SObject

from SFDCFW.Constant import SFDC_API_V


class SFDCFW(SObject):
    """Salesforce.com FrameWork."""

    def __init__(self,
                 username=None,
                 password=None,
                 security_token=None,
                 client_id=None,
                 client_secret=None,
                 version=SFDC_API_V,
                 domain='login',
                 wsdl=None,
                 metadata=False):
        """Constructor

        Args:
            username (str): The Salesforce user Username
            password (str): The Salesforce user Password
            security_token (str): The Salesforce user Security Token
            client_id (str): The Salesforce Connected App Consumer Key
            client_secret (str): The Salesforce Connected App Consumer Secret
            version (str): The Salesforce version of the Application Programming Interface
            domain (str): The common Salesforce domain for connection (login or test)
            wsdl (str): The path to the WSDL (Web Services Description Language) file
            metadata (bool): Whether or not this is for metadata
        """

        # Create an instance of Access object and login
        access = Access(username=username,
                        password=password,
                        security_token=security_token,
                        client_id=client_id,
                        client_secret=client_secret,
                        version=version,
                        domain=domain,
                        wsdl=wsdl,
                        metadata=metadata).login()

        # Unpack the tuple for session ID / access token and server URL / instance URL
        id_token, base_url = access

        # Parse the URL
        u = urlparse(base_url)
        self.base_url = f'{u.scheme}://{u.netloc}'

        # Create REST header
        self.header = {
            'Authorization': f'Bearer {id_token}',
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json'
        }


"""
SFDCFW.Rest.Rest
~~~~~~~~~~~~~~~~
"""

import json
from urllib.parse import urlparse


class Rest(object):
    """REST (REpresentational State Transfer) class."""

    def __init__(self, access):
        """Constructor

        Args:
            access (tuple): The Salesforce session ID / access token and
                server URL / instance URL tuple
        """

        # Unpack the tuple for session ID / access token and server URL / instance URL
        id_token, base_url = access
        
        # Parse the URL
        u = urlparse(base_url)
        self.base_url = f'{u.scheme}://{u.netloc}'

        # Create REST header
        self.header = {
            'Authorization': f'Bearer {id_token}',
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json'
        }

    def __getattr__(self, label):
        """Get Attribute Passed In.

        Args:
            label (str): The attribute passed in.

        Returns:
            A instance of the SObject class.
        """
        # Set the name / label
        self.label = label

        # Return the self instance
        return self


"""
SFDCFW.Test.TestRestSObjectAccount
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import json
import os
import random
import string
import unittest

from SFDCFW.Access import Access
from SFDCFW.SObject import SObject
from SFDCFW.Constant import TEST_DATA


def setUpModule():
    """Set Up Module"""
    pass


def tearDownModule():
    """Tear Down Module"""
    pass


class TestRestSObjectCreateAccount(unittest.TestCase):
    """
    Test the SObject module using REST (REpresentational State Transfer)
    with Create Account.
    """

    @classmethod
    def setUpClass(cls):
        """Prepare test setup class.

        Get the data from JSON (JavaScript Object Notation) file and
        login. Initialize any prerequisite.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the domain
        domain = cls.data['domain']

        # Get the REST Access user data for success login
        rest_access_user_success = cls.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_success['username'],
                        password=rest_access_user_success['password'],
                        security_token=rest_access_user_success['security_token'],
                        client_id=rest_access_user_success['consumer_key'],
                        client_secret=rest_access_user_success['consumer_secret'],
                        domain=domain).login()

        # Create an instance of SObject
        cls.sobject = SObject(access)


    def test_rest_sobject_create_account_success(self):
        """Test a success of REST (REpresentational State Transfer) SObject create Account.

        Generated Account Name to make a request for create. Should
        result in response with status code 204 No Content.
        """

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        account_id = self.sobject.Account.create(json.dumps(payload))

        # Variable `self.data` come from `cls.data`
        # Create the dictionary for Account data
        self.data['account'] = {}
        self.data['account']['rest_create_account_success'] = {}
        # Create or update the Account ID in the Test Data
        self.data['account']['rest_create_account_success']['id'] = account_id

        # Write the new Test Data to file
        with open(self.test_data_file, 'w') as f:
            json.dump(self.data, f)

        # Test to ensure Account ID is a string
        self.assertEqual(type(account_id), str)


    def test_rest_sobject_create_account_failure(self):
        """Test a failure of REST (REpresentational State Transfer) SObject create Account.

        Generated a random Account Description to make a request for
        create. Should result in response with status code 400 Bad
        Request.
        """

        # Create the payload
        payload = {
            # Generate a random Account Description
            'Description': f'Description-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        account_id = self.sobject.Account.create(json.dumps(payload))

        # Test to ensure Account ID is None
        self.assertEqual(account_id, None)


    @classmethod
    def tearDownClass(cls):
        """Prepare test teardown class.

        Clean up Test Data
        """

        # Check if Account data exist
        if 'account' in cls.data:
            # Get the REST create Account success data
            rest_create_account_success_id = cls.data['account']['rest_create_account_success']['id']

            # Make a request to delete the Account
            rest_create_account_success = cls.sobject.Account.delete(rest_create_account_success_id)

            # Check if the delete was successful
            if rest_create_account_success == 204:
                # Delete the Account entry from the Test Data
                del cls.data['account']

            # Write the new Test Data to file
            with open(cls.test_data_file, 'w') as f:
                json.dump(cls.data, f)


class TestRestSObjectReadAccount(unittest.TestCase):
    """
    Test the SObject module using REST (REpresentational State Transfer)
    with Read Account.
    """

    @classmethod
    def setUpClass(cls):
        """Prepare test setup class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the domain
        domain = cls.data['domain']

        # Get the REST Access user data for success login
        rest_access_user_success = cls.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_success['username'],
                        password=rest_access_user_success['password'],
                        security_token=rest_access_user_success['security_token'],
                        client_id=rest_access_user_success['consumer_key'],
                        client_secret=rest_access_user_success['consumer_secret'],
                        domain=domain).login()

        # Create an instance of SObject
        cls.sobject = SObject(access)

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.sobject.Account.create(json.dumps(payload))


    def test_rest_sobject_read_account_success(self):
        """Test a success for read Account.

        Make a request for read Account. Should result in response with
        a string and status code 200 OK.
        """

        # Make a request to read the Account
        # Get the return Account data
        account_data = self.sobject.Account.read(self.account_id)

        # Test to ensure Account data is a string
        self.assertEqual(type(account_data), str)


    def test_rest_sobject_read_account_failure(self):
        """Test a failure for read Account.

        Generate a fake random Account unique identifier (ID) and make
        a request for read Account. Should result in response with
        None (status code 404 Not Found).
        """

        # Generate a random Account unique identifier (ID)
        # The Account unique identifier (ID) will have invalid length
        account_id = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(8))

        # Make a request to read the Account
        # Get the return Account data
        account_data = self.sobject.Account.read(account_id)

        # Test to ensure Account data is None
        self.assertIsNone(account_data)


    @classmethod
    def tearDownClass(cls):
        """Prepare test teardown class.

        Clean up Test Data
        """

        # Make a request to delete the Account
        _ = cls.sobject.Account.delete(cls.account_id)


class TestRestSObjectUpdateAccount(unittest.TestCase):
    """
    Test the SObject module using REST (REpresentational State Transfer)
    with Update Account.
    """

    @classmethod
    def setUpClass(cls):
        """Prepare test setup class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the domain
        domain = cls.data['domain']

        # Get the REST Access user data for success login
        rest_access_user_success = cls.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_success['username'],
                        password=rest_access_user_success['password'],
                        security_token=rest_access_user_success['security_token'],
                        client_id=rest_access_user_success['consumer_key'],
                        client_secret=rest_access_user_success['consumer_secret'],
                        domain=domain).login()

        # Create an instance of SObject
        cls.sobject = SObject(access)

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.sobject.Account.create(json.dumps(payload))


    def test_rest_sobject_update_account_success(self):
        """Test a success for update Account.

        Generated Account Name to make a request for update. Should
        result in response with status code 204 No Content.
        """

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to update the Account
        account = self.sobject.Account.update(self.account_id, json.dumps(payload))

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(account, 204)


    def test_rest_sobject_update_account_failure(self):
        """Test a failure for update Account.

        Generate Account Name to make a request for update. Should
        result in response with status code 204 No Content.
        """

        # Generate a random Account unique identifier (ID)
        # The Account unique identifier (ID) will have invalid length
        account_id = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(8))

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to update the Account
        account = self.sobject.Account.update(account_id, json.dumps(payload))

        # Test to ensure HTTP status code is not 204 No Content
        self.assertNotEqual(account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test teardown class.

        Clean up Test Data
        """

        # Make a request to delete the Account
        _ = cls.sobject.Account.delete(cls.account_id)


class TestRestSObjectDeleteAccount(unittest.TestCase):
    """
    Test the SObject module using REST (REpresentational State Transfer)
    with Delete Account.
    """

    @classmethod
    def setUpClass(cls):
        """Prepare test setup class.

        Get the data from JSON (JavaScript Object Notation) file and
        login.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the domain
        domain = cls.data['domain']

        # Get the REST Access user data for success login
        rest_access_user_success = cls.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_success['username'],
                        password=rest_access_user_success['password'],
                        security_token=rest_access_user_success['security_token'],
                        client_id=rest_access_user_success['consumer_key'],
                        client_secret=rest_access_user_success['consumer_secret'],
                        domain=domain).login()

        # Create an instance of SObject
        cls.sobject = SObject(access)

        # Create the payload
        payload = {
            # Generate a random Account Name
            'Name': f'Account-{random.randrange(10000, 99999)}'
        }

        # Make a request to create the Account
        # Get the return unique identifier (ID)
        # The payload need to be serialized to JSON formatted str (json.dumps)
        cls.account_id = cls.sobject.Account.create(json.dumps(payload))


    def test_rest_sobject_delete_account_success(self):
        """Test a success for delete Account.

        Get the unique identifier (ID) of the Account from the Test Data
        to make a request for delete. Should result in response with
        status code 204 No Content.
        """

        # Make a request to delete the Account
        account = self.sobject.Account.delete(self.account_id)

        # Test to ensure HTTP status code is 204 No Content
        self.assertEqual(account, 204)


    def test_rest_sobject_delete_account_failure(self):
        """Test a failure for delete Account.

        Generate a random unique identifier (ID) of the Account to make
        a request for update. Should result in response with status
        code 204 No Content.
        """

        # Generate a random Account unique identifier (ID)
        # The Account unique identifier (ID) will have invalid length
        account_id = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(8))

        # Make a request to delete the Account
        account = self.sobject.Account.delete(account_id)

        # Test to ensure HTTP status code is not 204 No Content
        self.assertNotEqual(account, 204)


    @classmethod
    def tearDownClass(cls):
        """Prepare test teardown class.

        Clean up Test Data
        """

        # Make a request to delete the Account
        _ = cls.sobject.Account.delete(cls.account_id)


def suite():
    """Test Suite"""

    # Create the Unit Test Suite
    suite = unittest.TestSuite()

    # Add the Unit Test
    suite.addTest(TestRestSObjectCreateAccount())
    suite.addTest(TestRestSObjectReadAccount())
    suite.addTest(TestRestSObjectUpdateAccount())
    suite.addTest(TestRestSObjectDeleteAccount())

    # Return the Test Suite
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())


import json
import time
import sys

from SFDCFW.Rest.Rest import Rest

import pandas as pd

from SFDCFW.Constant import SFDC_API_V
from SFDCFW.Constant import HTTP_GET
from SFDCFW.Constant import HTTP_POST
from SFDCFW.Constant import HTTP_PATCH
from SFDCFW.Constant import HTTP_PUT

class Bulk(Rest):
    """Bulk API.
    """

    def create_job(self, operation, object_type):
        """Create a Bulk job

        Args:
            operation (str): The operation of the Bulk job
            object_type (str): The object type (API name) for the Bulk job

        Returns:
            A string formatted JSON for the HTTP response object
        """

        # Create the request relative URL
        relative_url = "/jobs/ingest"

        # Update the header
        self._header["Content-Type"] = "application/json"

        # Create payload
        payload = {
            "operation": operation,
            "object": object_type,
            "contentType": "CSV"
        }

        # Send the request
        r = self.send(HTTP_POST, relative_url, json.dumps(payload))

        return r.text


    def add_data(self, id, data):
        """Add data to the job

        Args:
            id (str): The ID of the Bulk job
            data (str): The data for the Bulk job, must be in CSV format

        Returns:
            A string of the response status code
        """

        # Create the data request relative URL
        relative_url = "/jobs/ingest/" + id + "/batches"

        # Update the header
        self._header["Content-Type"] = "text/csv"
        self._header["Accept"] = "application/json"

        # Send the data request
        r = self.send(HTTP_PUT, relative_url, data)

        # Return the response status code
        return r.status_code


    def close_job(self, id):
        """Close the job

        Args:
            id (str): The ID of the Bulk job

        Returns:
            A string formatted JSON for the HTTP response object
        """

        # Create the job request relative URL
        relative_url = "/jobs/ingest/" + id

        # Update the header
        self._header["Content-Type"] = "application/json"
        self._header.pop("Accept", None)

        # Create payload
        job_payload = {
            "state": "UploadComplete"
        }

        # Send the job request
        r = self.send(HTTP_PATCH, relative_url, json.dumps(job_payload))

        return r.text


    def check_status(self, id):
        """Check the status of the job

        Args:
            id (str): The ID of the Bulk job

        Returns:
            A string formatted JSON for the HTTP response object
        """

        # Create the job request relative URL
        relative_url = "/jobs/ingest/" + id

        # Update the header
        self._header.pop("Content-Type", None)
        self._header.pop("Accept", None)

        # Send the job request
        r = self.send(HTTP_GET, relative_url, None)

        return r.text


    def get_result(self, id):
        """Get the job result

        Args:
            id (str): The ID of the Bulk job

        Returns:
            A string formatted JSON for the HTTP response object
        """
        pass


    def create(self, object_type, data):
        """Create

        Args:
            data (str): The data to be created in CSV format (newline per entry)

        Returns:
            A string formatted JSON for the HTTP response object
        """

        # The specified operation for Bulk job
        operation = "insert"

        # Create the Bulk job
        bulk_job = self.create_job(operation, object_type)

        # Parse the Bulk job ID
        bulk_job_id = json.loads(bulk_job)["id"]
        # Create the data request relative URL
        data_relative_url = "/jobs/ingest/" + bulk_job_id + "/batches"
        # Addition to header
        self._header["Content-Type"] = "text/csv"
        self._header["Accept"] = "application/json"
        # Send the data request
        _ = self.send(HTTP_PUT, data_relative_url, data)

        # Create the job request relative URL
        job_relative_url = "/jobs/ingest/" + bulk_job_id
        # Update the header
        self._header["Content-Type"] = "application/json"
        # Create the job payload
        job_payload = {
            "state": "UploadComplete"
        }
        # Send the job request 
        job_r = self.send(HTTP_PATCH, job_relative_url, json.dumps(job_payload))

        return job_r.text


    def update(self, object_type, data, batch_limit=1000):
        """Update

        Args:
            object_type (str): The object type (API name) for the Bulk job
            data (list): A list of dictionary with the data to update, each
                dictionary should have the same "key"
                [
                    { "email": "alice@company.com.invalid" },
                    { "email": "bob@company.com.invalid" }
                ]

            batch_limit (int): The record size (count) of the Bulk job

        Returns:
            A string formatted JSON for the HTTP response object
        """

        print("Object Type: {}".format(object_type))
        print("Data: {}".format(data))

        # It is easier to extract data with pandas DataFrame
        # Create the DataFrame from data
        df = pd.DataFrame.from_dict(data)

        # The specified operation for Bulk job
        operation = "update"

        # Get the size of the object data
        object_data_size = len(data)
        # Calculate the number of Bulk job needed
        bulk_job_size = (object_data_size + (-object_data_size) % batch_limit) // batch_limit

        # Create a list for the Bulk job ID
        # This will be for each Bulk job
        bulk_job_id_list = []
        # Loop through and create the payload list for Bulk job
        for i in range(0, bulk_job_size):
            # Calculate the head (start) and tail (end) of each Bulk payload
            payload_head = i * batch_limit
            payload_tail = payload_head + batch_limit - 1

            # Create the Bulk job
            bulk_job_r = self.create_job(operation, object_type)

            # Parse the Bulk job ID
            bulk_job_id = json.loads(bulk_job_r)["id"]
            # Add the Bulk job ID to list
            bulk_job_id_list.append(bulk_job_id)

            # Add the data to the job and get the response
            add_data_r = self.add_data(bulk_job_id, df[payload_head : payload_tail].to_csv(index=False))

            if add_data_r == 201:
                # Upload successful, close the job
                _ = self.close_job(bulk_job_id)
            else:
                # Failure, something went wrong
                print("Failure, Status Code: {}".format(add_data_r))


    def empty_batch(self, object_type, condition=None, batch_limit=1000):
        """Empty

        Args:
            object_type (str): The object type (API name) for the Bulk job
            condition (str): The WHERE clause
            batch_limit (int): The record size (count) of the Bulk job
        """

        # Query all the data from the object
        # Query
        if condition:
            query = "SELECT Id FROM " + object_type + " WHERE " + condition
        else:
            query = "SELECT Id FROM " + object_type
        # Retrieve the object data
        object_data = self.query(query)
        # JSONify the object data
        object_data = json.loads(object_data)

        # Get the record count of the object data
        object_data_size = object_data["totalSize"]
        # Compare record count vs. batch limit, take the smaller of the two
        object_data_count = object_data_size if object_data_size < batch_limit else batch_limit

        # Create a list for the payload
        # This will be for each Bulk job
        payload_list = []
        # Loop through and create the payload list batch
        for i in range(0, object_data_size, batch_limit):
            # Build a string with the object data for the delete request body
            payload = "Id\n"
            # Loop through and build the payload
            for record in object_data["records"][i : i + object_data_count - 1]:
                payload += record["Id"] + "\n"

            # Add the payload batch to list
            payload_list.append(payload)

        # The specified operation for Bulk job
        operation = "delete"

        # Create a list for the Bulk job ID
        bulk_job_id_list = []
        # Loop through the payload batch and create the Bulk jobs
        for payload in payload_list:
            bulk_job_r = self.create_job(operation, object_type)

            # Parse the Bulk job ID
            bulk_job_id = json.loads(bulk_job_r)["id"]
            # Add the Bulk job ID to list
            bulk_job_id_list.append(bulk_job_id)

        # Align the Bulk job ID to the payload
        bulk_job_id_payload_list = zip(bulk_job_id_list, payload_list)

        # Loop through and add data (payload) to job (bulk_job_id)
        for bulk_job in bulk_job_id_payload_list:
            # Unpack the tuple
            bulk_job_id, payload = bulk_job
            # Add the data to the job and get the response
            add_data_r = self.add_data(bulk_job_id, payload)

            if add_data_r == 201:
                # Upload successful, close the job
                _ = self.close_job(bulk_job_id)
            else:
                # Failure, something went wrong
                print("Failure, Status Code: {}".format(add_data_r))


    def empty(self, object_type, condition=None, batch_size=1000):
        """Empty

        Args:
            object_type (str): The object type (API name) for the Bulk job
            condition (str): The WHERE clause
            batch_size (int): The record size of the Bulk job
        """

        # Query all the data from the object
        # Query
        if condition:
            query = "SELECT Id FROM " + object_type + " WHERE " + condition
        else:
            query = "SELECT Id FROM " + object_type
        # Retrieve the object data
        object_data = self.query(query)
        # JSONify the object data
        object_data = json.loads(object_data)

        # Build a string with the object data for the delete request body
        payload = "Id\n"
        for record in object_data["records"]:
            payload += record["Id"] + "\n"

        # The specified operation for Bulk job
        operation = "delete"

        # Create the Bulk job and get the response
        bulk_job_r = self.create_job(operation, object_type)

        # Parse the Bulk job ID
        bulk_job_id = json.loads(bulk_job_r)["id"]
        # Add the data to the job and get the response
        add_data_r = self.add_data(bulk_job_id, payload)

        if add_data_r == 201:
            # Upload successful, close the job
            _ = self.close_job(bulk_job_id)
        else:
            # Failure, something went wrong
            print("Failure, Status Code: {}".format(add_data_r))

        # Check the state of the job
        check_status_r = self.check_status(bulk_job_id)

        # Get the current state
        current_state = json.loads(check_status_r)["state"]

        # See if the job is done (need to use equality compare)
        while current_state != "JobComplete":
            # If job is not done, wait and check again
            print("Job State: {}".format(current_state))
            print("Waiting 10 Seconds...")
            time.sleep(10)
            check_status_r = self.check_status(bulk_job_id)
            current_state = json.loads(check_status_r)["state"]

        return check_status_r


class Email:
    """Email.
    """

    def mask(self, email):
        """Mask an email

        Args:
            email (str): The email address to mask

        Returns:
            A string of the masked email address
        """

        # Mask the email address
        mask_email = email + ".invalid"

        return mask_email


import json

from SFDCFW.Rest.SObject import SObject
from SFDCFW.Constant.Constant import HTTP_GET

class User(SObject):
    """User API
    """

    def assign_profile(self, username, profile_name):
        """Assign Profile

        Args:
            username (str): The Salesforce username to assign
            profile_name: (str): The Salesforce Profile name to assign

        Returns:
            True or False for the success or failure of the update
        """

        # User query
        user_query = "SELECT Id FROM User WHERE Username='" + username + "' and IsActive=true"
        # Create the User request URL
        user_relative_url = "/query/?q=" + user_query
        # Send the request for the User
        r = self.send(HTTP_GET, user_relative_url, None)

        if r.status_code == 200:
            # JSONify the return User data
            user_data = json.loads(r.text)

        # Profile query
        profile_query = "SELECT Id,Name FROM Profile WHERE Name='" + profile_name + "'"
        # Create the Profile request URL
        profile_relative_url = "/query/?q=" + profile_query
        # Send the request for the Profile
        r = self.send(HTTP_GET, profile_relative_url, None)

        if r.status_code == 200:
            # JSONify the return Profile data
            profile_data = json.loads(r.text)

        # Update User
        # Get the User ID
        user_id = user_data["records"][0]["Id"]
        # Get the Profile ID
        profile_id = profile_data["records"][0]["Id"]

        # Create payload
        payload = {
            "ProfileId": profile_id
        }

        # Execute the update
        r = self.User.update(user_id, payload)

        if r.status_code == 204:
            return True
        else:
            return False

    
    def assign_permission_set(self, username, permission_set_name):
        """Assign Permission Set

        Args:
            username (str): The Salesforce username to assign
            permission_set_name: (str): The Salesforce Permission Set name
                (API Name) to assign

        Returns:
            True or False for the success or failure of the update
        """
        
        # User query
        user_query = "SELECT Id FROM User WHERE Username='" + username + "' and IsActive=true"
        # Create the User request URL
        user_relative_url = "/query/?q=" + user_query
        # Send the request for the User
        r = self.send(HTTP_GET, user_relative_url, None)

        if r.status_code == 200:
            # JSONify the return User data
            user_data = json.loads(r.text)

        # Permission query
        permission_set_query = "SELECT Id FROM PermissionSet WHERE Name='" + permission_set_name + "'"
        # Create the Permission Set request URL
        permission_set_relative_url = "/query/?q=" + permission_set_query
        # Send the request for the Permission Set
        r = self.send(HTTP_GET, permission_set_relative_url, None)

        if r.status_code == 200:
            # JSONify the return Permission Set data
            permission_set_data = json.loads(r.text)

        # Update Permission Set Assignment
        # Get the User ID
        user_id = user_data["records"][0]["Id"]
        # Get the Profile ID
        permission_set_id = permission_set_data["records"][0]["Id"]

        # Create payload
        payload = {
            "AssigneeId": user_id,
            "PermissionSetId": permission_set_id
        }

        # Execute the update
        r = self.PermissionSetAssignment.create(payload)

        # print("Status Code: {}".format(r.status_code))
        # print("Message: {}".format(r.text))

        if r.status_code == 201:
            return True
        else:
            return False


import csv
from SFDCFW.Soap.Metadata import Metadata

class WorkflowRule:
    """Workflow Rule.
    """

    def __init__(self, access, wsdl):
        """Constructor.

        Args:
            access (tuple): The Salesforce session ID / access token and
                server URL / instance URL tuple.
            wsdl (str): The path to the WSDL file.
        """

        # Use the Metadata API
        self.metadata = Metadata(access, wsdl)

    def toggle_active(self, active, full_name=None):
        """(De)Activate specific Workflow Rule(s) for a given list of
           full name(s).

        Args:
            active (bool): A flag determine whether to activate or
                deactivate the Workflow Rule(s).
            full_name (list): A list of string(s) for the full name of
                the Workflow Rule to (de)activate. Will (de)activate all
                Workflow Rule(s) if no full name is passed.

        Returns:
            A complex list of dictionary for result of the Workflow
                Rule(s) (de)activation.
        """

        # Create the query
        query = [ { "type": "WorkflowRule" } ]

        # Get the list of Workflow Rule(s)
        workflow_rule_list = self.metadata.list_metadata(query)

        if full_name is not None:
            # Find full_name in workflow_rule_list
            # Loop through list of Workflow Rule(s) (workflow_rule_list) and find the requested (full_name) element
            workflow_rule_result = [workflow_rule for workflow_rule in workflow_rule_list if workflow_rule["fullName"] in full_name]
        else:
            workflow_rule_result = workflow_rule_list

        # Parse a list of full name from the Workflow Rule(s) result
        full_name_result = [workflow_rule["fullName"] for workflow_rule in workflow_rule_result]

        # Get the detail read result information from the full name
        read_result = self.metadata.read_metadata_thread(query[0]["type"], full_name_result)

        # Slice out the Workflow Rule(s) result that are (in)active
        if active:
            read_result[:] = [record for record in read_result if not record["active"]]
        elif not active:
            read_result[:] = [record for record in read_result if record["active"]]

        # Loop through the result and update the active status
        for record in read_result:
            if active:
                record["active"] = True
            elif not active:
                record["active"] = False

        # Update the metadata
        update_result = self.metadata.update_metadata_thread(read_result)

        # Return the update result
        return update_result


"""
SFDCFW.SObject
~~~~~~~~~~~~~~
"""

import json
import requests

from SFDCFW.Constant import SFDC_API_V
from SFDCFW.Rest.Rest import Rest

class SObject(Rest):
    """SObject class."""


    def create(self, payload):
        """Create SObject.

        Args:
            payload (dict): The required data for the SObject.

        Returns:
            A string for the unique identifier (ID) of the SObject.
        """
        
        # Create the request URL
        request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/sobjects/{self.label}'

        # Send the request
        r = requests.post(url=request_url,
                          headers=self.header,
                          data=payload)

        # Check the status code
        if r.status_code == 201:
            # Parse the unique identifier (ID) of the SObject
            sobject_id = json.loads(r.text)['id']
            # Return the unique identifier (ID) of the SObject
            return sobject_id

        # There was an error
        return None


    def read(self, id=None):
        """Read SObject.

        Args:
            id (str): The unique identifier (ID) of the SObject.

        Returns:
            A string formatted JSON for the request.
        """

        if id is not None:
            # Create the request URL with ID
            request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/sobjects/{self.label}/{id}'
        else:
            # Create the request URL without ID
            request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/sobjects/{self.label}'

        # Send the request
        r = requests.get(url=request_url,
                         headers=self.header)

        # Check the status code
        if r.status_code == 200:
            # Return the response text (message body)
            return r.text

        # There was an error
        return None


    def update(self, id, payload):
        """Update SObject.

        Args:
            id (str): The ID of the SObject.
            payload (dict): The updated data for the SObject.

        Returns:
            A HTTP Status Code (or None) of the response.
        """

        # Create the request URL
        request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/sobjects/{self.label}/{id}'

        # Send the request
        r = requests.patch(url=request_url,
                           headers=self.header,
                           data=payload)

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    def delete(self, id):
        """Delete SObject.

        Args:
            id (str): The ID of the SObject.

        Returns:
            A HTTP Status Code (or None) of the response.
        """

        # Create the request URL
        request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/sobjects/{self.label}/{id}'

        # Send the request
        r = requests.delete(url=request_url,
                            headers=self.header)

        # Check the status code
        if r.status_code == 204:
            # Return the status code
            return r.status_code

        # There was an error
        return None


    # def query(self, query):
    #     """Execute SOQL (Salesforce Object Query Language) Query

    #     Args:
    #         query (str): The SOQL (Salesforce Object Query Language) query

    #     Returns:
    #         A string formatted JSON for the query response
    #     """

    #     # Create the request URL
    #     request_url = "/query/?q=" + query

    #     # Send the request
    #     r = self.send(HTTP_GET, request_url, None)

    #     return r.text


    # def query_more(self, next_record_url):
    #     """Query Next Record Batch

    #     Args:
    #         next_record_url (str): The URL for the next batch of records

    #     Returns:
    #         A string formatted JSON for the query response
    #     """

    #     # Send the request
    #     r = self.send(HTTP_GET, next_record_url, None)

    #     return r.text


from setuptools import setup, find_packages

# Read the contents of the README.md file
from pathlib import Path
current_directory = Path(__file__).parent
long_description = (current_directory/'README.md').read_text()

setup(
    name = 'sfdcfw',
    version = '0.1.5',
    packages = find_packages(),

    # Dependency
    install_requires = [
        'pandas',
        'requests',
        'zeep',
    ],

    # Metadata
    author = 'Yan Kuang',
    author_email = 'YTKme@Outlook.com',
    description = 'Saleforce.com FrameWork.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.5',
    keywords = 'salesforce sfdc fw api',
)


"""
SFDCFW.Test.TestAccess
~~~~~~~~~~~~~~~~~~~~~~
"""

import json
import os
import unittest

from SFDCFW.Access import Access
from SFDCFW.Constant import TEST_DATA


def setUpModule():
    """Set Up Module"""
    pass


def tearDownModule():
    """Tear Down Module"""
    pass


class TestRestAccess(unittest.TestCase):
    """Test the Access module with REST (REpresentational State Transfer)."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the Test Data from JSON (JavaScript Object Notation) file.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.domain = cls.data['domain']


    def test_login_rest_success(self):
        """Test a success of REST (REpresentational State Transfer) login method.

        Get the success username and password (rest_access_user_success)
        data from the Test Data file and login. Should result in login
        method returning a tuple of the access token and instance URL.
        """

        # Get the REST Access user data for success login
        rest_access_user_success = self.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_success['username'],
                        password=rest_access_user_success['password'],
                        security_token=rest_access_user_success['security_token'],
                        client_id=rest_access_user_success['consumer_key'],
                        client_secret=rest_access_user_success['consumer_secret'],
                        domain=self.domain).login()

        # Test to ensure `access` is a tuple
        self.assertEqual(type(access), tuple)


    def test_login_rest_failure(self):
        """Test a failure of REST (REpresentational State Transfer) login method.

        Get the failure username and password (rest_access_user_failure)
        data from the Test Data file and login. Should result in login
        method returning a a `None` value.
        """

        # Get the REST Access user data for failure login
        rest_access_user_failure = self.data['user']['access_user_failure']

        # Create an instance of Access object and login
        access = Access(username=rest_access_user_failure['username'],
                        password=rest_access_user_failure['password'],
                        security_token=rest_access_user_failure['security_token'],
                        client_id=rest_access_user_failure['consumer_key'],
                        client_secret=rest_access_user_failure['consumer_secret'],
                        domain=self.domain).login()

        # Test to ensure `access` is `None`
        self.assertIsNone(access)


class TestSoapAccess(unittest.TestCase):
    """Test the Access module with SOAP (Simple Object Access Protocol)."""

    @classmethod
    def setUpClass(cls):
        """Prepare test set up class.

        Get the Test Data from JSON (JavaScript Object Notation) file.
        """

        # Get the current directory of the file
        current_directory = os.path.dirname(os.path.abspath(__file__))
        # Get the path of the Test Data file
        cls.test_data_file = os.path.join(current_directory, TEST_DATA)

        # Open the file for reading
        with open(cls.test_data_file, 'r') as f:
            cls.data = json.load(f)

        # Get the hostname from the Test Data
        cls.domain = cls.data['domain']

        # Get the WSDL (Web Service Definition Language) file path
        cls.enterprise_wsdl = os.path.join(current_directory, cls.data['enterprise_wsdl'])
        # cls.metadata_wsdl = os.path.join(current_directory, cls.data["metadata_wsdl"])


    def test_soap_login_success(self):
        """Test a success of SOAP (Simple Object Access Protocol) login method.

        Get the success username and password (soap_access_user_success)
        data from the Test Data file and login. Should result in login
        method returning a tuple of the session ID and (metadata) server
        URL.
        """

        # Get the SOAP Access user data for success login
        soap_access_user_success = self.data['user']['access_user_success']

        # Create an instance of Access object and login
        access = Access(username=soap_access_user_success['username'],
                        password=soap_access_user_success['password'],
                        security_token=soap_access_user_success['security_token'],
                        domain=self.domain,
                        wsdl=self.enterprise_wsdl,
                        metadata=True).login()

        # Test to ensure `access` is a tuple
        self.assertEqual(type(access), tuple)


    def test_soap_login_failure(self):
        """Test a failure of SOAP (Simple Object Access Protocol) login method.

        Get the failure username and password (soap_access_user_failure)
        data from the Test Data file and login. Should result in login
        method returning a `None` value.
        """

        # Get the SOAP Access user data for failure login
        soap_access_user_failure = self.data['user']['access_user_failure']

        # Create an instance of Access object and login
        access = Access(username=soap_access_user_failure['username'],
                        password=soap_access_user_failure['password'],
                        security_token=soap_access_user_failure['security_token'],
                        domain=self.domain,
                        wsdl=self.enterprise_wsdl,
                        metadata=True).login()

        # Test to ensure `access` is `None`
        self.assertIsNone(access)


def suite():
    """Test Suite"""

    # Create the Unit Test Suite
    suite = unittest.TestSuite()

    # Add the Unit Test
    suite.addTest(TestRestAccess)
    suite.addTest(TestSoapAccess)

    # Return the Test Suite
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())


"""
SFDCFW.Query
~~~~~~~~~~~~
"""

import json
import requests

from SFDCFW.Constant import SFDC_API_V
from SFDCFW.Rest.Rest import Rest


class Query(Rest):
    """Query class."""
    
    def query(self, query, more=False):
        """Execute SOQL (Salesforce Object Query Language) Query

        Retrieve query results, or more query results if the initial
        results are too large.

        Args:
            query (str): The SOQL (Salesforce Object Query Language) query
            more (bool): Whether to retrieve more query result

        Returns:
            A string formatted JSON for the query response
        """

        # Format the query
        if more:
            # If retrieving more query result
            # The `query` parameter is an identifier
            format_query = query
        else:
            # Replace ` ` with `+` valid query
            format_query = query.replace(' ', '+')

        # Create the request URL
        request_url = f'{self.base_url}/services/data/v{SFDC_API_V}/query/?q={format_query}'

        # Send the request
        r = requests.get(url=request_url,
                         headers=self.header)

        # Check the status code
        if r.status_code == 200:
            # Return the response text (message body)
            return r.text

        # There was an error
        return None

    
    def explain(self, query):
        """Get Performance Feedback

        Retrieve performance feedback on query (without executing),
        report, or list view.
        """
        pass